/*************************

Gazoot's 'dmpmaker' library, version 10, July 2001
gazoot@postmark.net

* This library dumps map information to a .dmp file. Very useful when you are
  betatesting your game and would like it to be preserved so you can edit it for
  the next version.

  There are two ways to use this library:

  1. The easy way: Using the Easy-to-use interface. This is the recommended way unless
     you want more automation than a html form can provide.

  The interface is available so you can select the flags you want to set using a html form.
  It's convenient because you won't have to bother about flag values and arguments to the
  acutal MakeDmp proc. To use the interface, put the following two lines in your code:

  #define DMP_EASY_INTERFACE
  MakeDmpInterface()

  Everything will then be taken care of. For more details how to set it up, see the examples
  below.

  * You need the html library if you would like to use the interface. It can be downloaded from
    this link: http://www.byond.com/script/dms.cgi?url=byond://Dantom.htmllib

  Your last settings in the form will be saved to a file called "dmp_form_values.sav".


  2. The harder way, that gives you a little more flexibility and automation:

  Use the proc 'MakeDmp(filename,flags,time_output)'. It's global, so anything can use it.

  Argument explanation:

  filename - A string, what you would like to save the file as.

  flags - how detailed you would like the saving to be. For example, if you know
          that all your turfs variables will be set at compile-time, you can omit
          the DMP_SAVE_TURFVARS flag.
          In general, use the DMP_SAVE_XXXVARS when you know that the atoms in
          question have user-made variables that are really important to preserve.
          (Flag explanations see below)

		  To use the flags, you have to OR them together, because they are binary flags.
		  If you want to use DMP_SAVE_OBJ and DMP_SAVE_OBJVARS, call MakeDmp like this:

		  MakeDmp("mymap.dmp",DMP_SAVE_OBJ|DMP_SAVE_OBJVARS)

          * Remember that for every flag you add, the longer it will take to save
            the map. A map with 60000 turfs and very few vars
            will take about 60 seconds to save with only DMP_SAVE_TURFVARS set.
            During the save time the game will freeze. The number of variables in the
            atoms can make quite a big impact on the saving speed.

          You don't have to worry about the hardcoded variables. (density, opacity etc...)
          They are saved automatically if different from the original value.

  time_output - If you want the map saving to be timed, set this to where you would like
  		  the output to go. world or src are two good examples. (Timing is in seconds.)


* Flag explanations:

	DMP_SAVE_OBJ			Include all objs in world when saving.
	DMP_SAVE_OBJ_CONTENTS	Saves the contens in all objs as well. Quite time-consuming.
	DMP_SAVE_NPC_MOB		Include all NPC mobs (mobs without client) in world when saving.
							Useful if you want all NonPlayerCharacters to remain where they are.
	DMP_SAVE_NPC_CONTENTS	Saves the contens in all NPCs as well. Quite time-consuming.
	DMP_SAVE_PC_MOB			Include all PC mobs (mobs with a client) in world when saving
							This one is rarely used.
	DMP_SAVE_PC_CONTENTS	Saves the contents in all PCs as well. Very time-consuming!
	DMP_SAVE_OBJVARS		Save all user-created obj variables.
	DMP_SAVE_MOBVARS		Save all user-created mob variables.
	DMP_SAVE_TURFVARS		Save all user-created turf variables.
	DMP_SAVE_AREAVARS   	Save all user-created area variables.

	DMP_SAVE_ALL			Save everything on map. Be careful, this can take a LONG time.
	DMP_SAVE_ALL_EXCEPT_PC	Save everything on map except Players (mobs with a client).


* What to do when saving is finished:

  If you have Dreammaker open with your game, and have saved the map with the same name as
  the original one (keep a backup just to be safe), just recompile the world and all the
  changes should be shown in the map editor. You can even keep the map open when recompiling.


* Three examples how to use this library:
  The first one is a verb that a GM (Game Master) can execute whenever he/she wants to
  dump the current world to a dmp:

mob
	GM
		verb
			makedmp()
				#define DMP_EASY_INTERFACE
				MakeDmpInterface()

  Next example is simply to exeute MakeDmp when the world deletes, and save the world
  to a map with the current date and time:

world
	Del()
		var/flags = DMP_SAVE_OBJ|DMP_SAVE_OBJVARS|DMP_SAVE_TURFVARS|DMP_SAVE_AREAVARS
		MakeDmp("[name] - [time2text(realtime,"DD-MMM-YYYY hhmm")].dmp",flags)
		..()

  Last example is when you are a GM and want a quick way to save the map with the same
  name and flags every time:

mob
	GM
		verb
			makedmp()
				var/flags = DMP_SAVE_OBJ|DMP_SAVE_OBJ_CONTENTS|DMP_SAVE_OBJVARS|\
							DMP_SAVE_TURFVARS|DMP_SAVE_AREAVARS
				MakeDmp("gamemap.dmp",flags,src)


* Over and underlays:

  The Dmpmaker uses a savefile called "dmp_over_underlays.sav" to save all over and
  underlays. If you want to save your over and underlays, you must add two lines at
  the end of world.New():

world
	New()
		... // Your stuff here
		#define DMP_SAVE_OVER_UNDERLAYS		// First line to add
		spawn() Dmp_Load_OU()				// Second line to add


* Combining this library with a savefile is recommended, since at this point, the library
  doesn't save object pointers. Your atoms will lose all pointers to
  other atoms when using DmpMaker. Hopefully I will create a savefile addition to this
  library for pointers as well.

TODO:	 	*** Object pointers are not supported yet! Your atoms will lose all pointers
			to other atoms when saving. ***
			Use only one char in the map if less than 52 combinations.
			More than 17576 unique locations aren't supported yet.


----------------------------
You are hereby granted to use this library, as long as you put my name
(Andreas "Gazoot" Sderlund) and a link to www.thiaoouba.com somewhere
in your game. (Credits page?)

Disclaimer: Use this on your own risk. By using this library you accept that I
            can't be held responsible for any damages. The library never writes
            to the game however, just to a textfile. Be careful about the filenames,
            keep a backup of your original map, and everything should be all right.
----------------------------
*************************/

var
	const
		DMP_SAVE_OBJ = 1
		DMP_SAVE_OBJ_CONTENTS = 2
		DMP_SAVE_NPC_MOB = 4
		DMP_SAVE_NPC_CONTENTS = 8
		DMP_SAVE_PC_MOB = 16
		DMP_SAVE_PC_CONTENTS = 32
		DMP_SAVE_OBJVARS = 64
		DMP_SAVE_MOBVARS = 128
		DMP_SAVE_TURFVARS = 256
		DMP_SAVE_AREAVARS = 512

		DMP_SAVE_ALL = 1023
		DMP_SAVE_ALL_EXCEPT_PC = 975

	list/remove_varlist = list("loc","x","y","z","type","parent_type",\
			"overlays","underlays","verbs","vars","contents",\
			"density","desc","dir","gender","icon","icon_state","layer",\
			"luminosity","name","opacity","suffix","tag","text","type","visibility")

	list/remove_moblist = list("ckey","client","group","key","sight")

proc
	MakeDmp(filename,flags,time_output)

		// Map specific
		var
			map = ""
			fullmap = ""
			chars

			// Lists
			objs_and_turfs[0]
			only_turfs[0]

			list/charlist = list("a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z",\
							"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z")
			charpos1 = 1
			charpos2 = 1
			charpos3 = 1
//			charpos4 = 1

		// Temporary variables
		var
			runtime
			tmpstr
			tmpchar
			turf/T
			area/A
			obj/O
			mob/M
			cache_string
			cache_letter

		if(time_output) runtime = world.realtime

		for(var/clvl=1;clvl<=world.maxz;clvl++)
			map += "(1,1,[clvl]) = {\"\n" // Add header for each level.
			for(var/y=world.maxy;y>=1;y--)
				for(var/x=1;x<=world.maxx;x++)
					T = locate(x,y,clvl)
					A = T.loc

					tmpstr = "("

					if(flags&DMP_SAVE_OBJ)
						for(O as obj in T)
							tmpstr += "[O.type]" + _getStdVars(O,flags) + ","

					if(flags&DMP_SAVE_NPC_MOB)
						for(M as mob in T)
							if(flags&DMP_SAVE_NPC_MOB&&!M.client)
								tmpstr += "[M.type]" + _getStdVars(M,flags) + ","
							else if(flags&DMP_SAVE_PC_MOB&&M.client)
								tmpstr += "[M.type]" + _getStdVars(M,flags) + ","

					tmpstr += "[T.type]" + _getStdVars(T,flags)
					tmpstr += ",[A.type]" + _getStdVars(A,flags) + ")"

					if(tmpstr==cache_string)
						map += cache_letter
					else if(T.contents.len)
						if(tmpstr in objs_and_turfs)
							cache_letter = objs_and_turfs[tmpstr]
							map += cache_letter
						else
							tmpchar = charlist[charpos1] + charlist[charpos2] + charlist[charpos3]
							if(++charpos3>52)
								charpos3 = 1; charpos2++
								if(charpos2>52)
									charpos2 = 1; charpos1++
							objs_and_turfs += tmpstr
							objs_and_turfs[tmpstr]=tmpchar
							map += tmpchar
							cache_letter = tmpchar
						cache_string = tmpstr
					else
						if(tmpstr in only_turfs)
							cache_letter = only_turfs[tmpstr]
							map += cache_letter
						else
							tmpchar = charlist[charpos1] + charlist[charpos2] + charlist[charpos3]
							if(++charpos3>52)
								charpos3 = 1; charpos2++
								if(charpos2>52)
									charpos2 = 1; charpos1++
							only_turfs += tmpstr
							only_turfs[tmpstr]=tmpchar
							map += tmpchar
							cache_letter = tmpchar
						cache_string = tmpstr

//					;

				map+="\n"
				if(length(map)>1000)
					fullmap += map
					map = ""

			map += "\"}\n\n" // Add footer for each level.

		// Flush map
		fullmap += map

		for(var/ch in only_turfs)
			chars += "\"[only_turfs[ch]]\" = [ch]\n"
		for(var/ch in objs_and_turfs)
			chars += "\"[objs_and_turfs[ch]]\" = [ch]\n"
		chars += "\n"

		fdel(filename)
		text2file(chars+fullmap,filename)

		if(time_output)
			time_output << "\n<b>DmpMaker finished!</b> Time to save map: <b>[(world.realtime-runtime)/10]</b> seconds.\n"

	_addVar(Avar,value)
		if(isnum(value))
			return "[Avar] = [value];"
		else if(istext(value))
			return "[Avar] = \"[value]\";"
		else if(isicon(value))
			return "[Avar] = '[value]';"
		else if(istype(value,/list))
			if(!value:len) return ""
			var/str = "list("
			for(var/a in value)
				str += _addListVar(a)
			if(length(str)>5)
				return "[Avar] = "+ copytext(str,1,length(str)) + ");"
			else
				return ""
		else
			return ""
//		else if(isloc(value))
//			;	// Savefile addition here in the (far) future...

	_addListVar(value)
		if(isnum(value))
			return "[value],"
		else if(istext(value))
			return "\"[value]\","
		else if(isicon(value))
			return "'[value]',"
		else if(istype(value,/list))
			if(!value:len) return ""
			var/str = "list("
			for(var/a in value)
				str += _addListVar(a)
			return copytext(str,1,length(str)) + ")"
		else
			return ""
//		else if(isloc(value))
//			;	// Savefile addition here in the (far) future...


	_addContents(list/objlist,flags)
		var/str = "newlist("
		for(var/A2 in objlist)
			str += "[A2:type]" + _getStdVars(A2,flags) + ","
		str = copytext(str,1,length(str))
		str += ")"
		if(length(str)>10)
			return str
		else
			return "list()"

	_getStdVars(atom/O, flags)
		var/tmps = "{"

		// Atom hard-coded variables
		if(O.density!=initial(O.density))
			tmps += "density = [O.density];"
		if(O.desc!=initial(O.desc))
			tmps += "desc = \"[O.desc]\";"
		if(O.dir!=initial(O.dir))
			tmps += "dir = [O.dir];"
		if(O.gender!=initial(O.gender))
			tmps += "gender = \"[O.gender]\";"
		if(O.icon!=initial(O.icon))
			tmps += "icon = '[O.icon]';"
		if(O.icon_state) //&&O.icon_state!=initial(O.icon_state))
			tmps += "icon_state = \"[O.icon_state]\";"
		if(O.layer!=initial(O.layer))
			tmps += "layer = [O.layer];"
		if(O.luminosity!=initial(O.luminosity))
			tmps += "luminosity = [O.luminosity];"
		if(O.name!=initial(O.name))
			tmps += "name = \"[O.name]\";"
		if(O.opacity!=initial(O.opacity))
			tmps += "opacity = [O.opacity];"
		if(O.suffix!=initial(O.suffix))
			tmps += "suffix = \"[O.suffix]\";"
		if(O.tag) //!=initial(O.tag))
			tmps += "tag = \"[O.tag]\";"
		if(O.text!=initial(O.text))
			tmps += "text = \"[O.text]\";"
		if(O.invisibility!=initial(O.invisibility))
			tmps += "invisibility = [O.invisibility];"
		if(O.visibility!=initial(O.visibility))
			tmps += "visibility = [O.visibility];"

		// Two special hard-coded variables only for mobs:
		if(ismob(O)&&flags&DMP_SAVE_MOBVARS)
			var/mob/M = O
			if(M.sight!=initial(M.sight))
				tmps += "sight = [M.sight];"
			if(M.client)
				tmps += "key = \"[M.key]\";"


		if(isobj(O)&&flags&DMP_SAVE_OBJVARS)
			#ifdef DMP_SAVE_OVER_UNDERLAYS
			tmps += Dmp_Save_OU(O)
			#endif
			for(var/V in O.vars-remove_varlist)
				if(!issaved(O.vars[V])) continue
				if(initial(O.vars[V])!=O.vars[V])
					tmps += _addVar(V,O.vars[V])
			if(O.contents.len&&flags&DMP_SAVE_OBJ_CONTENTS)
				tmps += "contents = " + _addContents(O.contents,flags) + ";"

		else if(ismob(O)&&flags&DMP_SAVE_MOBVARS)
			#ifdef DMP_SAVE_OVER_UNDERLAYS
			tmps += Dmp_Save_OU(O)
			#endif
			for(var/V in O.vars-remove_varlist-remove_moblist)
				if(!issaved(O.vars[V])) continue
				if(initial(O.vars[V])!=O.vars[V])
					tmps += _addVar(V,O.vars[V])
			if(O.contents.len)
				var/mob/M = O
				if(flags&DMP_SAVE_NPC_CONTENTS&&!M.client)
					tmps += "contents = " + _addContents(M.contents,flags) + ";"
				else if(flags&DMP_SAVE_PC_CONTENTS&&M.client)
					tmps += "contents = " + _addContents(M.contents,flags) + ";"

		else if(isturf(O)&&flags&DMP_SAVE_TURFVARS)
			#ifdef DMP_SAVE_OVER_UNDERLAYS
			tmps += Dmp_Save_OU(O)
			#endif
			for(var/V in O.vars-remove_varlist)
				if(!issaved(O.vars[V])) continue
				if(initial(O.vars[V])!=O.vars[V])
					tmps += _addVar(V,O.vars[V])

		else if(isarea(O)&&flags&DMP_SAVE_AREAVARS)
			#ifdef DMP_SAVE_OVER_UNDERLAYS
			tmps += Dmp_Save_OU(O)
			#endif
			for(var/V in O.vars-remove_varlist)
				if(!issaved(O.vars[V])) continue
				if(initial(O.vars[V])!=O.vars[V])
					tmps += _addVar(V,O.vars[V])

		if(lentext(tmps)>1)
			return copytext(tmps,1,length(tmps)) + "}"	// Could slow down...
		else
			return ""

// The over/underlays savefile 'hack'
#ifdef DMP_SAVE_OVER_UNDERLAYS
var
	dmp_idcount = 0

atom
	var
		dmp_id

proc
	Dmp_Load_OU()
		for(var/atom/O in world)
			if(!O.dmp_id) continue
			var/savefile/F = new("dmp_over_underlays.sav")
			F.cd = "/"
			F["d[++dmp_idcount]OL"] >> O.overlays
			F["d[dmp_idcount]UL"] >> O.underlays

	Dmp_Save_OU(atom/O)
		if(!O.overlays.len&&!O.underlays.len) return ""
		var/savefile/F = new("dmp_over_underlays.sav")
		F.cd = "/"
		F["d[++dmp_idcount]OL"] << O.overlays
		F["d[dmp_idcount]UL"] << O.underlays
		return "dmp_id = [dmp_idcount];"
#endif

#ifdef DMP_EASY_INTERFACE
#include <dantom\htmllib\htmllib.dm>

Form/interface

	form_reusable = 1
	submit = "Save map"

	var
		mapname

		flags
		flags_values = list("1","2","4","8","16","32","64","128","256","512","1023","975")
		flags_interface = CHECKLIST

		timeoutput
		timeoutput_value
		timeoutput_interface = CHECKBOX

	Initialize()
		var/savefile/F = new("dmp_form_values.sav")
		F.cd = "/"
		F["mapname"] >> mapname
		F["timeoutput"] >> timeoutput
		F["flags"] >> flags

	ProcessForm()
		var/mob/M = usr
		var/allflags = null
		if(!mapname)
			M << "<font color=red><b>You must enter a filename!</b></font>"
			return

		for(var/fl in flags)
			allflags |= text2num(fl)

		if(timeoutput) timeoutput = M

		MakeDmp(mapname,allflags,timeoutput)

		if(timeoutput) timeoutput = 1
		else timeoutput = 0

		var/savefile/F = new("dmp_form_values.sav")
		F.cd = "/"
		F["mapname"] << mapname
		F["timeoutput"] << timeoutput
		F["flags"] << flags

	HtmlLayout()
		return \
{"
<html>
<body bgcolor="#FAEA99" text="#42A17A">
<font face="Tahoma" size="2">
<h3>DmpMaker Interface</h3>
<b>Map filename:</b> [mapname] <font size="1">(Example: testmap.dmp)</font><br><br>
<table border=1>
    <tr>
        <td colspan=4><b>Turfs</b> and <b>areas</b> are automatically saved on the map.</td>
    </tr>
    <tr>
        <td colspan=4><b>Objs:</b></td>
    </tr>
    <tr>
        <td>Save objs</td>
        <td colspan=3>Save contents of objs*</td>
    </tr>
    <tr>
        <td align="center">[flags["1"]]</td>
        <td align="center">[flags["2"]]</td>
        <td align="center">&nbsp;</td>
        <td>&nbsp;</td>
    </tr>
    <tr>
        <td colspan=4><b>Mobs:</b> (a NPC is a mob without a client, PC has a client.)</td>
    </tr>
    <tr>
        <td>Save NPC</td>
        <td>Save NPC contents*</td>
        <td>Save PC</td>
        <td>Save PC contents*</td>
    </tr>
    <tr>
        <td align="center">[flags["4"]]</td>
        <td align="center">[flags["8"]]</td>
        <td align="center">[flags["16"]]</td>
        <td align="center">[flags["32"]]</td>
    </tr>
	<tr>
	 <td colspan=4>[flags["64"]] Save all user-created obj variables.</td>
	</tr>
	<tr>
	 <td colspan=4>[flags["128"]] Save all user-created mob variables.</td>
	</tr>
	<tr>
	 <td colspan=4>[flags["256"]] Save all user-created turf variables.</td>
	</tr>
	<tr>
	 <td colspan=4>[flags["512"]] Save all user-created area variables.</td>
	</tr>
	<tr>
	 <td colspan=4>[flags["1023"]] Save everything on map. Be careful, this can take a LONG time.</td>
	</tr>
	<tr>
	 <td colspan=4>[flags["975"]] Save everything on map except Players (mobs with a client).</td>
	</tr>
</table>
<br><font size="1">* = Could be time-consuming and use the whole CPU for a while</font><br><br>
<b>If you want to time the map saving, check this box:</b> [timeoutput]<br>
With this checked, you will receive a confirmation when the saving is complete.
<br><br>[submit]
<br><br><hr>
<font size="1">DmpMaker interface, by <a href=mailto:gazoot@postmark.net>Gazoot</a>, July 2001.</font><br>
</body>
</html>
"}

proc
	MakeDmpInterface(mob/target=usr)
		var/Form/interface/form = new()
		form.DisplayForm(target)

#endif
